#! /usr/bin/env python2

#############################################################
#                                                           #
#                                                           #
#         arc v1.6                                          #
#                                                           #
#  (c) 2003 F. Kauff                                        #
#           J. Miadlikowska                                 #
#           F. Lutzoni                                      #
#  Dept. if Biology, Duke University                        #
#  USA                                                      #
#                                                           #
# contact (bugs, critics, etc):                             #
# fkauff@duke.edu                                           #
#                                                           #
# 03/10/2004                                                #
#                                                           #
#############################################################

import sys,getopt

PRECISION=2
AMINOS='ARNDCQEGHILKMFPSTWYVBZ'
STATELIST='0123456789BDEFHIJKLMOPQRSUVWXYZbdefhijklmnopqrsuvwxyz@#$%&!^_<>'
NSTATES=len(STATELIST)                     
TRIPLE={'TTT':'F','TTC':'F','TTA':'L','TTG':'L','TCT':'S','TCC':'S','TCA':'S','TCG':'S',
'TAT':'Y','TAC':'Y','TAA':'?','TAG':'?','TGT':'C','TGC':'C','TGA':'?','TGG':'W','CTT':'L',
'CTC':'L','CTA':'L','CTG':'L','CCT':'P','CCC':'P','CCA':'P','CCG':'P','CAT':'H','CAC':'H',
'CAA':'Q','CAG':'Q','CGT':'R','CGC':'R','CGA':'R','CGG':'R','ATT':'I','ATC':'I','ATA':'I',
'ATG':'M','ACT':'T','ACC':'T','ACA':'T','ACG':'T','AAT':'N','AAC':'N','AAA':'K','AAG':'K',
'AGT':'S','AGC':'S','AGA':'R','AGG':'R','GTT':'V','GTC':'V','GTA':'V','GTG':'V','GCT':'A',
'GCC':'A','GCA':'A','GCG':'A','GAT':'D','GAC':'D','GAA':'E','GAG':'E','GGT':'G','GGC':'G',
'GGA':'G','GGG':'G'}

class Matrix:

    def __init__(self, filename,mode=0):
        file=open(filename,'rU')

        # get ntax and nchar from line 1
        words=file.readline().split()
        self.ntax=int(words[0])
        self.nchar=int(words[1])
        self.nambigs=int(words[2])
        self.gap=words[3][0]
        self.unknown=words[4][0]
        if len(words)>5 and mode==0:
            self.nmotifs=int(words[5])
        else:
            self.nmotifs=0
        if mode==0:
            self.nnewchars=23+self.nmotifs
        else:
            self.nnewchars=1+len(AMINOS)
        self.newchars=[]
        for a in range(self.nambigs):
            newcharline=[]
            for s in range(self.ntax):
                newcharline.append(['.']*self.nnewchars)
            self.newchars.append(newcharline)
        # get ambig regions from line 2
        words=file.readline().split()
        self.ambigspos=[]
        for ambig in words:
            region=ambig.split('-')
            self.ambigspos.append((int(region[0]),int(region[1])))
        # get motifs from line 3
        if self.nmotifs>0:
            self.motifs=file.readline().upper().split()
            if self.nmotifs!=len(self.motifs):
                raise Exception, "Motif number discrepancy!"
        else:
            self.motifs=[]
        # get sequences from following lines
        self.seqs=[]
        self.taxnames=[]
        for i in range(self.ntax):
            try:
                line=file.readline()
                seq=line.split()
                self.seqs.append(seq[1].upper())
                self.taxnames.append(seq[0])
            except:
                raise
                #sys.exit('Syntax error in '+filename+':\n'+line)        
        self.ambigs=[]
        for ambig in self.ambigspos:
            ambiglist=[]
            for seq in self.seqs:
                #print seq[ambig[0]-1:ambig[1]],
                ambigreg=seq[ambig[0]-1:ambig[1]].replace(self.gap,'')
                # print seq[ambig[0]-1:ambig[1]],
                if mode==0:
                    ambiglist.append(ambigreg)
                else:
                    aa=''
                    for trips in range(0,len(ambigreg)-2,3):
                        aa+=triple(ambigreg[trips:trips+3])
                    # print aa
                    ambiglist.append(aa)
            #print
            self.ambigs.append(ambiglist)
        file.close()

def print_charfreqblock(matrix,cfb,nr,prot,file=sys.stdout):
    print >> file
    print >> file, 'Table of character values for ambiguous region no. %d (%d-%d)' % (nr+1,matrix.ambigspos[nr][0],matrix.ambigspos[nr][1])
    if prot==1:
        print >> file, '                     A    R    N    D    C    Q    E    G    H    I    L    K    M    F    P    S    T    W    Y    V    B    Z'
    print >> file, '              ',
    for i in range(len(cfb[0])):
        print >> file, '%2d  ' % (i+1),
    print >> file
    for list,taxname in zip(cfb,matrix.taxnames):
        print >> file, '%-12s' % taxname,
        for freq in list:
            print >> file, '%4s' % str(freq),
        print >> file 

def print_codematrix(matrix,cb,nr,prot,file=sys.stdout):
    print >> file 
    print >> file, '[Characters for ambiguous region no. %d (%d-%d)]' % (nr+1,matrix.ambigspos[nr][0],matrix.ambigspos[nr][1])
    if prot==1:
        print >> file, '[             ARNDCQEGHILKMFPSTWYVBZ]'
    print >> file, '[            ',
    for i in range(matrix.nnewchars/10):
        print >> file, '       ',i+1,
    print >> file, ']'
    codestring=''
    for i in range(matrix.nnewchars):         
        codestring=codestring+str((i+1) % 10)
    print >> file, '[            %s]' % codestring
    for list,taxname in zip(cb,matrix.taxnames):
        codestring=''
        for code in list:
            codestring=codestring+code
        print >> file, '%-12s' % taxname,codestring

def print_docu():
    print "ARC: Ambiguous Regions Coding"
    print "(following Miadlikowska et al. 2003, xxxx.)"
    print "extended for use with nucleotides and amino acids"
    print "Usage: %s -f filename [-p]" % (sys.argv[0].split('/')[-1])
    print "Switch -p is optional and indicates protein mode, otherwise nucleotide mode is assumed"
    print """

A) Nucleotide mode:

Input file format:
(line 1) <no. of taxa> <no. of characters> <no. of ambig. regions> <char. for gaps> <char. for missing data> <no. of motifs>
(line 2) <start-end of ambig. region1> ... <start-end of ambig. region n>
(line 3) <motif 1> ... <motif n>  # omit line 3 completely when not checking for motifs
(line 4) <name of taxon 1> <sequence of taxon 1>
...
(line n) <name of taxon n> <sequence of taxon n>

Examples:
3 10 2 - ? 2                 3 10 2 - ? 0
3-4 7-9                      3-4 7-9
acttt ggt                    dog acttgcc-gt
dog acttgc-ggt               cat gct??cctgt
cat gct??cctgt               fly gctta-tttg
fly gctta-tttg

Output files are <filename>.val with actual character values (frequencies etc.) and <filename>.code with coded character table for (e.g.) PAUP*
Characters are (1) sequence length, (2-5) A,C,G,T frequencies, (6-15) AA,AC,AG,AT,CC,CG,CT,GG,GT,TT frequencies,
(16-19) AA,CC,GG,TT pair frequencies, (20-23) A,C,G,T distribution, (24-x) motif presence [1] or absence [0]

A)  Amino acid mode

Input file format:
(line 1) <no. of taxa> <no. of characters> <no. of ambig. regions> <char. for gaps> <char. for missing data> 
(line 2) <start-end of ambig. region1> ... <start-end of ambig. region n>
(line 3) <name of taxon 1> <sequence of taxon 1>
...
(line n) <name of taxon n> <sequence of taxon n>

When used in amino acid mode, the ambiguous regions are translated into amino acids before processing. Only the length of the amino
acid sequence and the relative frequencies of each amino acid will be calculated and coded.
Output files are <filename>.val with actual character values (frequencies etc.) and <filename>.code with coded character table for (e.g.) PAUP*
Characters are (1) amino acid sequence length, (2-23) (A,R,N,D,C,Q,E,G,H,I,L,K,M,F,P,S,T,W,Y,V,B,Z) frequencies

For more information see documentation.

(c) 03/2003 F. Kauff (fkauff@duke.edu) / J. Miadlikowska / F.Lutzoni 
"""

def freqstring(freq,p):
    fs=str(round(freq,p))
    fs=fs[:fs.find('.')+3]
    dot=fs.find('.')
    if dot==-1:
        fs=fs+'.00'
    elif len(fs[dot:])==2:
        fs=fs+'0'
    if not len(fs[dot:])==3:
        print "Freqstring failed:",fs
        raise Exception 
    return fs

def basecount(seq,base):
    return seq.count(base)

def basefreq(seq,base):
    if len(seq)==0:
        return '0.00'
    else:
        return freqstring(float(basecount(seq,base))/len(seq),PRECISION)

def paircount(seq,pair):
    pc=0
    pair2=pair[1]+pair[0]
    for pos in range(len(seq)-1):
        if seq[pos:pos+2]==pair or seq[pos:pos+2]==pair2: pc=pc+1
    return pc
	

def pairfreq(seq,pair):
    if len(seq)>1:
        return freqstring(float(paircount(seq,pair))/(len(seq)-1),PRECISION) 
    else:
        return '0.00'

def motifcount(seq,motif,mode=0):
    mc=0
    for pos in range(len(seq)-len(motif)+1):
        if seq[pos:pos+len(motif)]==motif: mc=mc+1
    if mode==0:
        if mc>0:
            return '1'
        else:
            return '0'
    else:
        return str(mc)

def doublebase(seq,base):
    db=0
    for pos in range(len(seq)-1):
        if seq[pos]==base and seq[pos+1]==base: db=db+1
    return db

def doublebasefreq(seq,base):
    if basecount(seq,base)>0:
        return freqstring(float(doublebase(seq,base))/basecount(seq,base),PRECISION)
    else:
        return '0.00'

def distrbase(seq,base):
    dib=0
    for pos in range(len(seq)-1):
        if seq[pos]==base and not seq[pos+1]==base:
            dib=dib+1
    if seq[-1]!=base: dib=dib-1
    return dib

def distrbasefreq(seq,base):
    if basecount(seq,base)>0:
        return freqstring(float(distrbase(seq,base))/basecount(seq,base),PRECISION) 
    else:
        return '0.00'

def triple(code):
	code=code.upper()
	if len(code)!=3:
		return '@'
	elif TRIPLE.has_key(code):
		return TRIPLE[code]
	elif code=='---':
		return '-'
	elif code==':::':
		return ':'
	elif code[0] in 'ACGT?' and code[1] in 'ACGT?' and code[2] in 'ACGT?':
		return '?'
	else:
		return 'X'

### main

prot=0
filename=''

#We need Python2.3 at least!
v1,v2=sys.version_info[:2]
if v1<2 or (v1==2 and v2<3):
    print sys.argv[0],'requires Python 2.3 or above. Check www.python.org for update.'
    sys.exit(0)
try:
    opts,left=getopt.getopt(sys.argv[1:],'f:ph',['help'])
    for o,a in opts:
        if o=='-f':
            filename=a
        elif o=='-p':
            prot=1
        elif o=='-h' or o=='--help':
            print_docu()
            sys.exit()
except getopt.GetoptError:
    print_docu()
    sys.exit(0)
if filename=='':
    print_docu()
    sys.exit(0)

freqtableout=filename+'.val'
codetableout=filename+'.code'
matrix=Matrix(filename,prot)

print 'Input file           :', filename
print 'No. of taxa          :', matrix.ntax
print 'Mode                 :',
if prot==1:
    print 'amino acids'
else:
    print 'nucleotides'
print 'Sequences length     :', matrix.nchar
print 'No. of ambig. regions:', matrix.nambigs
print 'Gap                  :', matrix.gap
print 'Missing data         :', matrix.unknown
print 'No. of motifs        :', matrix.nmotifs,
if matrix.nmotifs>0:
    print matrix.motifs
elif prot==1:
    print '(amino acid mode does not support motifs)'

fto=open(freqtableout,'w')
cto=open(codetableout,'w')
fto.write('Input file: ')
fto.write(filename)
fto.write('\n')
cto.write('[Input file: ')
cto.write(filename)
cto.write(']\n\n[Caution: character table is case sensitive when having more than 30 character states]\n')

ambignr=0
maxcoding=0
for ambigblock in matrix.ambigs:
    charfreqblock=[]
    seqnr=0
    for ambig in ambigblock:
        # print matrix.taxnames[seqnr],ambig    
        charfreqlist=[]
        if '?' not in ambig:
            charfreqlist.append(len(ambig))
        else:
            charfreqlist.append('?')            # 1  single nucleotides
        if prot==0:
            if 'N' not in ambig and '?' not in ambig:
                charfreqlist.append(basefreq(ambig,'A'))        # 2
                charfreqlist.append(basefreq(ambig,'C'))        # 3
                charfreqlist.append(basefreq(ambig,'G'))        # 4
                charfreqlist.append(basefreq(ambig,'T'))        # 5
                charfreqlist.append(pairfreq(ambig,'AA'))         # 6  nucleotide pairs
                charfreqlist.append(pairfreq(ambig,'AC'))         # 7
                charfreqlist.append(pairfreq(ambig,'AG'))         # 8
                charfreqlist.append(pairfreq(ambig,'AT'))         # 9
                charfreqlist.append(pairfreq(ambig,'CC'))         # 10
                charfreqlist.append(pairfreq(ambig,'CG'))         # 11
                charfreqlist.append(pairfreq(ambig,'CT'))         # 12
                charfreqlist.append(pairfreq(ambig,'GG'))         # 13
                charfreqlist.append(pairfreq(ambig,'GT'))         # 14
                charfreqlist.append(pairfreq(ambig,'TT'))         # 15
                charfreqlist.append(doublebasefreq(ambig,'A'))       # 16   contiguous pairs
                charfreqlist.append(doublebasefreq(ambig,'C'))       # 17
                charfreqlist.append(doublebasefreq(ambig,'G'))       # 18
                charfreqlist.append(doublebasefreq(ambig,'T'))       # 19
                charfreqlist.append(distrbasefreq(ambig,'A'))       # 20    base ditribution
                charfreqlist.append(distrbasefreq(ambig,'C'))       # 21
                charfreqlist.append(distrbasefreq(ambig,'G'))       # 22
                charfreqlist.append(distrbasefreq(ambig,'T'))        # 24
                for motif in matrix.motifs:            	# 25+x
                    charfreqlist.append(motifcount(ambig,motif))
            else:
                for i in range(matrix.nnewchars-1):
                    charfreqlist.append('?')
        else:        # protein mode: check only amino acid frequencies
            if 'X' not in ambig and '?' not in ambig:
                for aa in AMINOS:
                    charfreqlist.append(basefreq(ambig,aa))
            else:
                for i in range(matrix.nnewchars-1):
                    charfreqlist.append('?')
                
        charfreqblock.append(charfreqlist)
        seqnr=seqnr+1

    # code new characters
    print_charfreqblock(matrix,charfreqblock,ambignr,prot,fto)
    for nc in range(matrix.nnewchars-matrix.nmotifs):    # code character values, except motif presence/absence
        coding=0
        for seqc in range(matrix.ntax):
            if charfreqblock[seqc][nc]=='?':
                matrix.newchars[ambignr][seqc][nc]='?'
            elif matrix.newchars[ambignr][seqc][nc]=='.':
                matrix.newchars[ambignr][seqc][nc]=STATELIST[coding]
                for seqc2 in range(seqc+1,matrix.ntax):
                    if charfreqblock[seqc][nc]==charfreqblock[seqc2][nc]:
                        matrix.newchars[ambignr][seqc2][nc]=STATELIST[coding]
                coding=coding+1
                if coding>maxcoding: maxcoding=coding
                if coding==len(STATELIST):
                    cto.close()
                    fto.close()
                    print "Error while coding ambiguous region %d, Taxon '%s', character %d:" % (ambignr+1,matrix.taxnames[seqc],nc+1)
                    raise Exception, 'Too many character states!' 
    for nc in range(matrix.nnewchars-matrix.nmotifs,matrix.nnewchars):    # motif presence/absence must be copied, not coded
        for seqc in range(matrix.ntax):
            matrix.newchars[ambignr][seqc][nc]=charfreqblock[seqc][nc]    # copy presence/absence of motif
    print_codematrix(matrix,matrix.newchars[ambignr],ambignr,prot,cto)
    ambignr=ambignr+1

cto.close()
fto.close()
print "Character states used:",STATELIST[:maxcoding]
print "\nTable with character values written to file '%s'" % freqtableout
print "Table with character states written to file '%s'" % codetableout 

